###############################################################################
# assign_category function determines the heat alert category that a set of 
# temperature values belongs to, based on high-low-high thresholds. The 
# category value is equal to the sum of the total degrees by which the daytime
# highs and the overnight lows deviated below the candidate high-low-high
# thresholds. If a temperature is >= to its corresponding threshold, it would 
# contribute 0 degrees to the total. If a forecast is 1 degree lowel than its
# threshold, it would contribute 1 degree to the total, and so on. The total
# sum of these differences is used to assign dates to the heat alert categories
# as follows: Category 0 = 0 degrees; Category 1 = 1 degree; Category 2 = 2 
# degrees; and Category 3 = 3 degrees or more.
#
# high_temp1 = first daytime high temperature for the current date & scenario
# high_temp2 = second daytime high temperature for the current date & scenario
# low_temp = overnight low temperature for the current date & scenario
# high_threshold = high threshold
# low_threshold = low threshold
#
# output = character vector of the assigned categories (i.e. "Category 2")
#
# example: if high_threshold == 30 and low_threshold == 15, and 
# high_temp1 == 29, high_temp2 == 30, and low_temp == 16, then only high_temp1 
# is below the high_threshold and the total difference is 1. Thus, this would 
# be Category 1.
# 
###############################################################################

assign_category <- function(high_temp1, high_temp2, low_temp, 
                            high_threshold, low_threshold){
  
  diffs <- data.frame(diff1 = high_threshold - high_temp1,
                      diff2 = high_threshold - high_temp2,
                      diff3 = low_threshold - low_temp) %>%
    mutate_all(funs(ifelse(. < 0, 0, .))) %>%
    mutate(sum_diffs = rowSums(.),
           category = case_when(
             sum_diffs == 0 ~ "Category 0",
             sum_diffs == 1 ~ "Category 1",
             sum_diffs == 2 ~ "Category 2",
             sum_diffs > 2 ~ "Category 3"
           ))
  
  diffs$category
}

###############################################################################
# model_the_data function cleans the data and runs it through a time-series
# model. 
#
# dataset = data frame with the heat alert category and response variable
# response_var = character name of the response variable for the regression
# equation
#
# output = GLM object
# 
###############################################################################

model_the_data <- function(dataset, response_var){
  
  # clean the dataset
  data_clean <- dataset[!is.na(dataset$category),]
  data_clean[["category"]] <- factor(data_clean[["category"]], 
                                     levels = c("Category 3", "Category 2", 
                                                "Category 1", "Category 0"))
  
  # summarise the category variable
  categories <- data.frame(table(data_clean[, "category"]),
                           stringsAsFactors = F)
  names(categories) <- c("category", "n")
  
  # create character object of the regression equation
  f <- paste0(response_var, " ~ category + ns(dayt, df = 6)")
  
  # run the model
  if(nrow(categories) > 1){
    output <- tryCatch(do.call("glm", list(as.formula(f), 
                                           data = as.name("data_clean"), 
                                           family = "poisson")),
                       error = function(err) NULL)
  } else {
    output <- NULL
  }
  
  output
}

###############################################################################
# create_results_df function extracts the data from the models produced by the 
# model_the_data function in preparation for visualizing the data.
#
# high_threshold = high threshold value
# low_threshold = low threshold value
# model_data = GLM from the model_the_data function
#
# output = data frame rate ratios and confidence intervals from the model
# 
###############################################################################

create_results_df <- function(high_threshold, low_threshold, model_data){
  
  if(is.null(model_data)){
    results <- data.frame("high_threshold" = high_threshold,
                          "low_threshold" = low_threshold,
                          "category" = "Category 2",
                          "coefficient" = NA,
                          "SE" = NA,
                          "RR" = NA,
                          "LCI" = NA,
                          "UCI" = NA,
                          row.names = NULL)
  } else{
    # get the exponentiated coefficients and confidence intervals
    coeffs <- model_data$coefficients
    rrs <- exp(coeffs[grep("category", names(coeffs))])
    coeffs <- coeffs[grep("category", names(coeffs))]
    stderrors <- summary(model_data)$coefficients[, "Std. Error"]
    stderrors <- stderrors[grep("category", names(stderrors))]
    confints <- tryCatch(exp(confint(model_data, level = 0.95)), 
                         error = function(err) matrix(NA, length(coeffs), 2))
    if(all(is.na(confints))){
      confints <- matrix(NA, length(coeffs), 2)
      row.names(confints) <- names(coeffs)
    } else{confints <- confints[grep("category", row.names(confints)), , drop = F]}
    
    # put them in a data frame
    results <- data.frame("high_threshold" = high_threshold,
                          "low_threshold" = low_threshold,
                          "category" = substr(row.names(confints), 
                                              nchar("category")+1, 
                                              nchar("category")+10),
                          "coefficient" = coeffs,
                          "SE" = stderrors,
                          "RR" = rrs,
                          "LCI" = confints[, 1],
                          "UCI" = confints[, 2],
                          row.names = NULL)
  }
  
  results$category <- factor(results$category, 
                             levels = c("Category 3", "Category 2", 
                                        "Category 1", "Category 0"))
  
  results
}

###############################################################################
# perform_metaanalysis function performs the meta-analysis for all models
# from one heat alert area and one combination of high-low-high thresholds.
#
# dataset = data frame including columns named: category, coefficient, SE
# haa = character name of the heat alert area
#
# output = data frame of the meta-analysis results
# 
###############################################################################

perform_metaanalysis <- function(dataset, haa){
  
  # Make the category column a factor *without* Category 3 as a level
  dataset$category <- factor(dataset$category,
                             levels = c("Category 2", "Category 1", "Category 0"))
  
  ma <- metagen(TE = coefficient,
                seTE = SE,
                byvar = category,
                data = dataset)
  
  ma_results <- data.frame("heat_alert_area" = haa,
                           "category" = c("Category 2", "Category 1", "Category 0"),
                           "estimate" = exp(summary(ma)$within.random$TE),
                           "lci" = exp(summary(ma)$within.random$lower),
                           "uci" = exp(summary(ma)$within.random$upper),
                           stringsAsFactors = F)
  
  ma_results
}